package display

import (
	"reflect"
	"sort"
	"strings"
	"sync"
	"time"
)

const (
	sysConfigVersion  = "1.0"
	userConfigVersion = "1.0"
)

type SysRootConfig struct {
	mu       sync.Mutex
	Version  string
	Config   SysConfig
	UpdateAt string
}

func (c *SysRootConfig) copyFrom(newSysConfig *SysRootConfig) {
	c.mu.Lock()

	c.Version = newSysConfig.Version
	c.Config = newSysConfig.Config
	c.UpdateAt = newSysConfig.UpdateAt

	c.mu.Unlock()
}

// SysConfig v1
type SysConfig struct {
	DisplayMode  byte
	Screens      map[string]*SysScreenConfig
	ScaleFactors map[string]float64 // 缩放比例
	FillModes    map[string]string  // key 是特殊的 fillMode Key
	Cache        SysCache
}

type SysCache struct {
	BuiltinMonitor string
	ConnectTime    map[string]time.Time
}

// UserConfig v1
type UserConfig struct {
	Version string
	Screens map[string]UserScreenConfig
}

func (c *UserConfig) fix() {
	for _, screenConfig := range c.Screens {
		screenConfig.fix()
	}
}

type UserScreenConfig map[string]*UserMonitorModeConfig

const (
	KeyModeMirror        = "Mirror"
	KeyModeExtend        = "Extend"
	KeyModeOnlyOnePrefix = "OnlyOne-"
	KeySingle            = "Single"
)

type UserMonitorModeConfig struct {
	ColorTemperatureMode   int32
	ColorTemperatureManual int32
	// 以后如果有必要
	// Monitors UserMonitorConfigs
}

func (c *UserMonitorModeConfig) fix() {
	if !isValidColorTempMode(c.ColorTemperatureMode) {
		c.ColorTemperatureMode = defaultTemperatureMode
	}
	if !isValidColorTempValue(c.ColorTemperatureManual) {
		c.ColorTemperatureManual = defaultTemperatureManual
	}
}

func getDefaultUserMonitorModeConfig() *UserMonitorModeConfig {
	return &UserMonitorModeConfig{
		ColorTemperatureMode:   defaultTemperatureMode,
		ColorTemperatureManual: defaultTemperatureManual,
	}
}

func (m *Manager) getUserScreenConfig() UserScreenConfig {
	id := m.getMonitorsId()
	screenCfg := m.userConfig.Screens[id]
	if screenCfg == nil {
		if m.userConfig.Screens == nil {
			m.userConfig.Screens = make(map[string]UserScreenConfig)
		}
		screenCfg = UserScreenConfig{}
		m.userConfig.Screens[id] = screenCfg
	}
	return screenCfg
}

func (usc UserScreenConfig) getMonitorModeConfig(mode byte, uuid string) (cfg *UserMonitorModeConfig) {
	switch mode {
	case DisplayModeMirror:
		return usc[KeyModeMirror]
	case DisplayModeExtend:
		return usc[KeyModeExtend]
	case DisplayModeOnlyOne:
		return usc[KeyModeOnlyOnePrefix+uuid]
	}
	return nil
}

func (usc UserScreenConfig) setMonitorModeConfig(mode byte, uuid string, cfg *UserMonitorModeConfig) {
	switch mode {
	case DisplayModeMirror:
		usc[KeyModeMirror] = cfg
	case DisplayModeExtend:
		usc[KeyModeExtend] = cfg
	case DisplayModeOnlyOne:
		usc[KeyModeOnlyOnePrefix+uuid] = cfg
	}
}

func (usc UserScreenConfig) fix() {
	for _, config := range usc {
		config.fix()
	}
}

// SysScreenConfig 系统级屏幕配置
// NOTE: Single 可以看作是特殊的显示模式，和 Mirror,Extend 等模式共用 SysMonitorModeConfig 结构可以保持设计上的统一，不必在乎里面有 Monitors
type SysScreenConfig struct {
	Mirror      *SysMonitorModeConfig            `json:",omitempty"`
	Extend      *SysMonitorModeConfig            `json:",omitempty"`
	Single      *SysMonitorModeConfig            `json:",omitempty"`
	OnlyOneMap  map[string]*SysMonitorModeConfig `json:",omitempty"`
	OnlyOneUuid string                           `json:",omitempty"`
}

func (c *SysScreenConfig) fix() {
	if c.Mirror != nil {
		c.Mirror.fix()
	}
	if c.Extend != nil {
		c.Extend.fix()
	}
	if c.Single != nil {
		c.Single.fix()
	}
	for uuid, config := range c.OnlyOneMap {
		if config != nil {
			config.fix()
		} else {
			delete(c.OnlyOneMap, uuid)
		}
	}
}

type SysMonitorModeConfig struct {
	Monitors SysMonitorConfigs
}

func (c *SysMonitorModeConfig) fix() {
	for _, monitor := range c.Monitors {
		monitor.fix()
	}
}

type SysMonitorConfig struct {
	UUID        string
	Name        string
	Enabled     bool
	X           int16
	Y           int16
	Width       uint16
	Height      uint16
	Rotation    uint16
	Reflect     uint16
	RefreshRate float64
	Brightness  float64
	Primary     bool
}

func (c *SysMonitorConfig) fix() {
	// c.Enable 为 false，但 c.Primary 为 true 的，错误情况
	if !c.Enabled && c.Primary {
		c.Primary = false
	}
	if !isValidBrightness(c.Brightness) {
		c.Brightness = 1
	}
}

func (c *SysMonitorConfig) modify(changes monitorChanges) {
	for name, value := range changes {
		var ok bool
		switch name {
		case monitorPropX:
			c.X, ok = value.(int16)
		case monitorPropY:
			c.Y, ok = value.(int16)
		case monitorPropWidth:
			c.Width, ok = value.(uint16)
		case monitorPropHeight:
			c.Height, ok = value.(uint16)
		case monitorPropRotation:
			c.Rotation, ok = value.(uint16)
		case monitorPropRefreshRate:
			c.RefreshRate, ok = value.(float64)
		case monitorPropEnabled:
			c.Enabled, ok = value.(bool)
		default:
			ok = true
			logger.Warningf("invalid monitor property name, uuid: %v, name: %v", c.UUID, name)
		}

		if !ok {
			logger.Warningf("failed to set value, uuid: %v, name: %v, value: %v", c.UUID, name, value)
		} else {
			logger.Debugf("config %v set value %v = %v", c.UUID, name, value)
		}
	}
}

func isValidBrightness(value float64) bool {
	// 不含 0
	return value > 0 && value <= 1
}

type SysMonitorConfigs []*SysMonitorConfig

func (cfgs SysMonitorConfigs) clone() SysMonitorConfigs {
	if cfgs == nil {
		return nil
	}
	result := make(SysMonitorConfigs, len(cfgs))
	for i, config := range cfgs {
		configCp := *config
		result[i] = &configCp
	}
	return result
}

func (s *SysScreenConfig) getSingleMonitorConfigs() SysMonitorConfigs {
	if s.Single == nil {
		return nil
	}
	return s.Single.Monitors
}

func (s *SysScreenConfig) getMonitorConfigs(mode uint8, uuid string) SysMonitorConfigs {
	switch mode {
	case DisplayModeMirror:
		if s.Mirror == nil {
			return nil
		}
		return s.Mirror.Monitors

	case DisplayModeExtend:
		if s.Extend == nil {
			return nil
		}
		return s.Extend.Monitors

	case DisplayModeOnlyOne:
		if uuid == "" {
			return nil
		}
		if s.OnlyOneMap[uuid] == nil {
			return nil
		}
		return s.OnlyOneMap[uuid].Monitors
	}

	return nil
}

func (s *SysScreenConfig) setSingleMonitorConfigs(configs SysMonitorConfigs) {
	if s.Single == nil {
		s.Single = &SysMonitorModeConfig{}
	}
	s.Single.Monitors = configs
}

func (s *SysScreenConfig) setMonitorConfigs(mode uint8, uuid string, configs SysMonitorConfigs) {
	switch mode {
	case DisplayModeMirror:
		if s.Mirror == nil {
			s.Mirror = &SysMonitorModeConfig{}
		}
		s.Mirror.Monitors = configs

	case DisplayModeExtend:
		if s.Extend == nil {
			s.Extend = &SysMonitorModeConfig{}
		}
		s.Extend.Monitors = configs

	case DisplayModeOnlyOne:
		s.setMonitorConfigsOnlyOne(uuid, configs)
	}
}

func (s *SysScreenConfig) setMonitorConfigsOnlyOne(uuid string, configs SysMonitorConfigs) {
	if uuid == "" {
		return
	}
	if s.OnlyOneMap == nil {
		s.OnlyOneMap = make(map[string]*SysMonitorModeConfig)
	}

	if s.OnlyOneMap[uuid] == nil {
		s.OnlyOneMap[uuid] = &SysMonitorModeConfig{}
	}

	if len(configs) > 1 {
		// 去除非使能的 monitor
		var tmpCfg *SysMonitorConfig
		for _, config := range configs {
			if config.Enabled {
				tmpCfg = config
				break
			}
		}
		if tmpCfg != nil {
			configs = SysMonitorConfigs{tmpCfg}
		}
	}

	s.OnlyOneMap[uuid].Monitors = configs
}

func (cfgs SysMonitorConfigs) getByUuid(uuid string) *SysMonitorConfig {
	for _, mc := range cfgs {
		if mc.UUID == uuid {
			return mc
		}
	}
	return nil
}

func (cfgs SysMonitorConfigs) setPrimary(uuid string) {
	for _, mc := range cfgs {
		if mc.UUID == uuid {
			mc.Primary = true
		} else {
			mc.Primary = false
		}
	}
}

// cfgs 和 otherCfgs 之间是否仅亮度不同
// 前置条件：cfgs 和 otherCfgs 不相同
func (cfgs SysMonitorConfigs) onlyBrNotEq(otherCfgs SysMonitorConfigs) bool {
	if len(cfgs) != len(otherCfgs) {
		return false
	}
	// 除了亮度设置为 0， 其他字段都复制
	partCpCfgs := func(cfgs SysMonitorConfigs) SysMonitorConfigs {
		copyCfgs := make(SysMonitorConfigs, len(cfgs))
		for i, cfg := range cfgs {
			cpCfg := &SysMonitorConfig{}
			*cpCfg = *cfg
			cpCfg.Brightness = 0
			copyCfgs[i] = cpCfg
		}
		return copyCfgs
	}

	c1 := partCpCfgs(cfgs)
	c2 := partCpCfgs(otherCfgs)
	// 把亮度都安全的设置为0, 如果 c1 和 c2 是相同的，则可以说明是仅亮度不同。
	if reflect.DeepEqual(c1, c2) {
		return true
	}
	return false
}

func (cfgs SysMonitorConfigs) sort() {
	sort.Slice(cfgs, func(i, j int) bool {
		return strings.Compare(cfgs[i].UUID, cfgs[j].UUID) < 0
	})
}
